@@ -1,10 +1,9 @@
 #![doc = include_str!("../README.md")]
-use anyhow::Result;
-use chrono::{DateTime, Local};
+use chrono::{DateTime, Duration, Local};
 use clap::Parser;
 use colored::Colorize;
-use reqwest::StatusCode;
 use scraper::{Html, Selector};
+use std::error;
 use std::fs::OpenOptions;
 use std::io::Write;
 use url::Url;
@@ -47,15 +46,23 @@ impl CheckError {
     }
 }
 
+/// get absolute link
+fn get_abs_link(href: &str, referer: Url) -> String {
+    // if link is already an absolute link, use as is
+    if href.starts_with("http") || href.starts_with("www") {
+        href.to_string()
+    // convert relative to absolute link
+    } else {
+        referer.join(href).unwrap().to_string()
+    }
+}
+
 /// get all links from one url
-fn get_links(url: &Url) -> Result<Vec<Link>> {
+fn get_links(url: &Url) -> Result<Vec<Link>, Box<dyn error::Error>> {
     let res = reqwest::blocking::get(url.as_ref())?;
 
     // current url, even if redirected
-    let full_url = &res.url().clone();
-
-    // get domain using Url::parse
-    let domain = format!("{}://{}", url.scheme(), url.host_str().unwrap());
+    let referer: &Url = &res.url().clone();
 
     // get and parse the html
     let html = res.text()?;
@@ -64,26 +71,17 @@ fn get_links(url: &Url) -> Result<Vec<Link>> {
 
     let mut links = vec![];
 
-    // go through all a href elements and modify value as necessary, push to links
+    // go through all "a" elements
     for element in document.select(&selector) {
+        // if it has an href value, modify that value as necessary and push to links
         if let Some(mut v) = element.value().attr("href") {
             v = v.trim();
             // don't add mailto links or anchor links
             if v.starts_with("mailto") || v.starts_with('#') {
                 continue;
-            }
-            // if link is full url, add it as is
-            if v.starts_with("http") {
-                let link = Link::new(full_url.clone(), Url::parse(v)?);
-                links.push(link);
-            // if starts with /, just need the domain added
-            } else if v.starts_with('/') {
-                let link = Link::new(full_url.clone(), Url::parse(&format!("{}{}", domain, v))?);
-                links.push(link);
-            // if neither of those, this is a relative link starting from full url
             } else {
-                let link = Link::new(full_url.clone(), Url::parse(&format!("{}{}", full_url, v))?);
-                links.push(link);
+                let link = Url::parse(&get_abs_link(v, referer.clone()))?;
+                links.push(Link::new(referer.clone(), link));
             }
         }
     }
@@ -92,9 +90,9 @@ fn get_links(url: &Url) -> Result<Vec<Link>> {
 }
 
 /// check validity of individual link
-fn check_link(url: &Url) -> Result<bool> {
+fn check_link(url: &Url) -> Result<bool, reqwest::Error> {
     let res = reqwest::blocking::get(url.as_ref())?;
-    Ok(res.status() != StatusCode::NOT_FOUND)
+    Ok(res.status() != reqwest::StatusCode::NOT_FOUND)
 }
 
 fn main() {
@@ -131,6 +129,7 @@ fn main() {
     println!("\nChecking links found in {:?}", url.to_string());
 
     // start the check
+    let start: DateTime<Local> = Local::now();
     let mut unchecked_links: Vec<Link> = vec![Link::new(url.clone(), url)];
     let mut valid_links: Vec<Link> = vec![];
     let mut invalid_links: Vec<Link> = vec![];
@@ -248,7 +247,26 @@ fn main() {
         // }
     }
 
-    println!("\nChecked {} links.", count);
+    let elapsed: Duration = Local::now() - start;
+
+    let human_elapsed: String;
+    if elapsed.num_hours() > 0 {
+        human_elapsed = format!(
+            "{} hours, {} minutes",
+            elapsed.num_hours(),
+            elapsed.num_minutes()
+        );
+    } else if elapsed.num_minutes() > 0 {
+        human_elapsed = format!(
+            "{} minutes, {} seconds",
+            elapsed.num_minutes(),
+            elapsed.num_seconds()
+        );
+    } else {
+        human_elapsed = format!("{} seconds", elapsed.num_seconds());
+    }
+
+    println!("\nChecked {} links in {}.", count, human_elapsed,);
 
     // process results
     if invalid_links.is_empty() {
@@ -304,3 +322,81 @@ fn main() {
         }
     }
 }
+
+#[test]
+fn abs_link_domain_referer_no_trailing_slash() {
+    let referer = Url::parse("https://kdwarn.dev").unwrap();
+    assert_eq!(
+        get_abs_link("https://kdwarn.dev/hello/", referer),
+        "https://kdwarn.dev/hello/"
+    )
+}
+
+#[test]
+fn abs_link_domain_referer_trailing_slash() {
+    let referer = Url::parse("https://kdwarn.dev/").unwrap();
+    assert_eq!(
+        get_abs_link("https://kdwarn.dev/hello/", referer),
+        "https://kdwarn.dev/hello/"
+    )
+}
+
+#[test]
+fn abs_link_www_domain_referer() {
+    let referer = Url::parse("https://kdwarn.dev").unwrap();
+    assert_eq!(
+        get_abs_link("www.kdwarn.dev/hello/", referer),
+        "www.kdwarn.dev/hello/"
+    )
+}
+
+#[test]
+fn rel_link_with_slash_domain_referer_no_trailing_slash() {
+    let referer = Url::parse("https://kdwarn.dev").unwrap();
+    assert_eq!(get_abs_link("/hello", referer), "https://kdwarn.dev/hello")
+}
+
+#[test]
+fn rel_link_with_slash_domain_referer_trailing_slash() {
+    let referer = Url::parse("https://kdwarn.dev/").unwrap();
+    assert_eq!(get_abs_link("/hello", referer), "https://kdwarn.dev/hello")
+}
+
+#[test]
+fn rel_link_no_slash_domain_referer_no_trailing_slash() {
+    let referer = Url::parse("https://kdwarn.dev").unwrap();
+    assert_eq!(get_abs_link("hello", referer), "https://kdwarn.dev/hello")
+}
+
+#[test]
+fn rel_link_no_slash_domain_referer_trailing_slash() {
+    let referer = Url::parse("https://kdwarn.dev/").unwrap();
+    assert_eq!(get_abs_link("hello", referer), "https://kdwarn.dev/hello")
+}
+
+#[test]
+fn rel_link_no_slash_referer_with_subdir_trailing_slash() {
+    let referer = Url::parse("https://kdwarn.dev/hello/").unwrap();
+    assert_eq!(
+        get_abs_link("kris", referer),
+        "https://kdwarn.dev/hello/kris"
+    )
+}
+
+#[test]
+fn rel_link_no_slash_referer_with_subdir_no_trailing_slash() {
+    let referer = Url::parse("https://kdwarn.dev/hello/there").unwrap();
+    assert_eq!(
+        get_abs_link("kris", referer),
+        "https://kdwarn.dev/hello/kris"
+    )
+}
+
+#[test]
+fn rel_link_from_filename() {
+    let referer = Url::parse("https://kdwarn.dev/hello/old.htm").unwrap();
+    assert_eq!(
+        get_abs_link("kris", referer),
+        "https://kdwarn.dev/hello/kris"
+    )
+}
